-----------------------------------------------------------------------
-- "Natural Order" comparison & sorting algorithm
-- v2.1 (01/13/2006)
-- Andy Serpa (ac@onehorseshy.com)
--
-- What it does:
-- Will sort a sequence of character strings in
-- the order a person would normally put
-- them in if the strings contain numbers.
-- Especially useful for sorting lists of
-- filenames with numbers in them.
--
-- Example:
-- {"1A","Item10","Item1A","Item100","Item20","Item1","100","Item10B","10A"}
--
-- A regular sort will sort this sequence as:
-- {"100","10A","1A","Item1","Item10","Item100","Item10B","Item1A","Item20"}
--
-- A natural order sort will sort as follows:
-- {"1A","10A","100","Item1","Item1A","Item10","Item10B","Item20","Item100"}
--
-- The regular sort orders everything alphabetically, including the embedded
-- numbers, so for instance it puts "10" before "2" because the starting
-- 'letter' -- the digit '1' -- comes before the digit '2'.  Natural ordering
-- compares numbers as numbers, ordering them numerically. (2 is compared to 10,
-- not to 1.)  This comparison only occurs when the start of each number
-- "lines up" with the other and everything preceding them is equal.
--
-- For example, "Item2" would be placed before "Item10", but not before "I10".
--
--
-- The algorithm only takes into account "unbroken" integers within the
-- character strings -- it makes no allowance for decimal points (although it
-- will still work correctly as long as there are an equal number of digits
-- to the right of the decimal point in each number being compared).
--
--
-- Included in this version are four global functions:
--
-- nat_compare()
-- nat_compare_str()
--
-- nat_sort()
-- nat_sort_str()
--
--
-- These routines work just like compare() & sort(), except the
-- comparisons are done using the natural order algorithm.  The
-- nat_sort() & nat_sort_str() functions simply use the custom_sort()
-- library function and invoke nat_compare() or nat_compare_str() as
-- the comparision functions.
--
-- nat_compare() will compare any two Euphoria objects, &
-- nat_sort() will sort any sequence of any depth & composition.
-- [Like sort(), only the top-level elements are being sorted.]
--
-- nat_compare_str() & nat_sort_str() are meant to be used when
-- it is known that all elements to be compared will be
-- character strings (or else it will crash).
-- They are about as twice as fast.
--
--
-- Usage:
--              Just like compare():
--                              i = nat_compare(x1,x2)
--                              (i =  1  if x1 > x2,
--                               i =  0  if x1 = x2,
--               i = -1  if x1 < x2)
--
--              And just like sort():
--                              s = nat_sort(s)
--                              (sorted in ascending natural order)
--
--
-- Notes on version 2.0:
-- Although I'm still using the custom_sort() routine, which causes some
-- redundancy in processing the strings, the algorithm has been completely
-- re-written for this version taking a different approach.
--
-- The basic functions nat_compare() & nat_sort() are about twice as fast
-- as the old versions, and the new functions nat_compare_str() & nat_sort_str()
-- are about twice as fast as that.
--
-- Of course, they're all still quite a bit slower than regular sorting,
-- but whaddya gonna do?  (Normal sorting is blazingly fast, really.)
--
--
-- v2.1 Bug fix:
--      changed one declaration line to avoid integer overflow when comparing strings with
--  long strings of digits

include sort.e

-- compares any two objects using the natural order algorithm
global function nat_compare(object x1, object x2)
integer x1pt, x2pt, x1len, x2len, number, comp, string
atom x1num, x2num
object x1char, x2char

        if sequence(x1) and sequence(x2) then

                x1len = length(x1)
                x2len = length(x2)

                -- if both are strings
                -- inline string type-checking for speed
                string = 1
                for i = 1 to x1len do
                        x1char = x1[i]
                        if not integer(x1char)
                        or x1char > 255
                        or (x1char < ' ' and not find(x1char,"\t\n\r")) then
                                -- not a string
                                string = 0
                                exit
                        end if
                end for

                if string then
                        string = 1
                        for i = 1 to x2len do
                                x2char = x2[i]
                                if not integer(x2char)
                                or x2char > 255
                                or (x2char < ' ' and not find(x2char,"\t\n\r")) then
                                        -- not a string
                                        string = 0
                                        exit
                                end if
                        end for
                end if

                if not string then
                        -- recurse
                        for i = 1 to x1len do
                                if i > x2len then
                                        return 1
                                end if
                                comp = nat_compare(x1[i],x2[i])
                                if comp != 0 then
                                        return comp
                                end if
                        end for
                        return compare(x1len,x2len)
                end if

                -- both are strings
                x1pt = 1
                x2pt = 1

                -- number flag
                number = 0

                while x1pt <= x1len do

                        if number then
                                -- build numbers
                                x1num = x1char - '0'
                                while x1pt < x1len do
                                        x1pt += 1
                                        x1char = x1[x1pt]
                                        if (x1char >= '0' and x1char <= '9') then
                                                x1num = (x1num * 10) + (x1char - '0')
                                        else
                                                x1pt -= 1
                                                exit
                                        end if
                                end while

                                x2num = x2char - '0'
                                while x2pt < x2len do
                                        x2pt += 1
                                        x2char = x2[x2pt]
                                        if (x2char >= '0' and x2char <= '9') then
                                                x2num = (x2num * 10) + (x2char - '0')
                                        else
                                                x2pt -= 1
                                                exit
                                        end if
                                end while

                                -- compare as atoms
                                comp = compare(x1num,x2num)
                                if comp != 0 then
                                        return comp
                                end if

                                -- still equal, keep going
                                number = 0
                                x1pt += 1
                                x2pt += 1

                        else

                                if x2pt > x2len then
                                        -- equal up to point x2 stops
                                        -- x1 is longer, therefore greater
                                        return 1
                                end if

                                x1char = x1[x1pt]
                                x2char = x2[x2pt]

                                if
                                (x1char >= '0' and x1char <= '9')
                                and
                                (x2char >= '0' and x2char <= '9')
                                then
                                        -- first digit of number found, set flag
                                        number = 1
                                else
                                        comp = compare(x1char,x2char)
                                        if comp != 0 then
                                                return comp
                                        end if
                                        x1pt += 1
                                        x2pt += 1
                                end if
                        end if
                end while
                -- equal up to point x1 stops, longer (if either) is greater
                return compare(x1len,x2len)
        else
                -- one or both top-level elements are atoms
                -- compare normally
                return compare(x1,x2)
        end if

end function




-- compares two character strings using natural ordering
-- (eliminates string type checking, recursion, etc. = faster)
global function nat_compare_str(sequence x1, sequence x2)
integer x1pt, x2pt, x1len, x2len, number, comp
atom x1num, x2num
object x1char, x2char

        x1len = length(x1)
        x2len = length(x2)

        x1pt = 1
        x2pt = 1

        -- number flag
        number = 0

        while x1pt <= x1len do

                if number then
                        -- build numbers
                        x1num = x1char - '0'
                        while x1pt < x1len do
                                x1pt += 1
                                x1char = x1[x1pt]
                                if (x1char >= '0' and x1char <= '9') then
                                        x1num = (x1num * 10) + (x1char - '0')
                                else
                                        x1pt -= 1
                                        exit
                                end if
                        end while

                        x2num = x2char - '0'
                        while x2pt < x2len do
                                x2pt += 1
                                x2char = x2[x2pt]
                                if (x2char >= '0' and x2char <= '9') then
                                        x2num = (x2num * 10) + (x2char - '0')
                                else
                                        x2pt -= 1
                                        exit
                                end if
                        end while

                        -- compare as atoms
                        comp = compare(x1num,x2num)
                        if comp != 0 then
                                return comp
                        end if

                        -- still equal, keep going
                        number = 0
                        x1pt += 1
                        x2pt += 1

                else

                        if x2pt > x2len then
                                -- equal up to point x2 stops
                                -- x1 is longer, therefore greater
                                return 1
                        end if

                        x1char = x1[x1pt]
                        x2char = x2[x2pt]

                        if
                        (x1char >= '0' and x1char <= '9')
                        and
                        (x2char >= '0' and x2char <= '9')
                        then
                                -- first digit of number found, set flag
                                number = 1
                        else
                                comp = compare(x1char,x2char)
                                if comp != 0 then
                                        return comp
                                end if
                                x1pt += 1
                                x2pt += 1
                        end if
                end if
        end while
        -- equal up to point x1 stops, longer (if either) is greater
        return compare(x1len,x2len)
end function


-- for sorting any sequence of any depth
global function nat_sort(sequence s)
        return custom_sort(routine_id("nat_compare"),s)
end function

-- for sorting sequences where all elements are character strings
global function nat_sort_str(sequence s)
        return custom_sort(routine_id("nat_compare_str"),s)
end function

global function myEZsort(sequence seq1,sequence seq2)--this only sorts 2 sequences takes                                       
   sequence newseq,syms,names                        --in 2 seqs of the same length
    newseq={}
    names={}
    syms={}
    for x=1  to length(seq1) do
        newseq=append(newseq,{seq1[x],seq2[x]})
    end for
    newseq=nat_sort(newseq)
    for x=1 to length(seq1) do
    --  printf(1,"%d %s\n",{x,newseq[x][1]})
        syms=append(syms,newseq[x][1])
        names=append(names,newseq[x][2])
    end for
   return {syms,names}
end function 


global function myEZsort3(sequence seq1,sequence seq2,sequence seq3)--this only sorts 3 sequences takes                                    
   sequence newseq,syms,names,other                      --in 3 seqs of the same length
    newseq={}
    names={}
    syms={}
    other={}
     -- printf(1,"length(s1)=%d length(s2)=%d length(s3)=%d\n",{length(seq1),length(seq2),length(seq3)})
    for x=1  to length(seq1) do
        newseq=append(newseq,{seq1[x],seq2[x],seq3[x]})
    end for
    newseq=nat_sort(newseq)
    for x=1 to length(seq1) do
    --  printf(1,"%d %s\n",{x,newseq[x][1]})
        syms=append(syms,newseq[x][1])
        names=append(names,newseq[x][2])
        other=append(other,newseq[x][3])
    end for
   return {syms,names,other}
end function 
